;; This example sets up a mode 1 screen. WIth a background a sprite that moves over
;; it (it will flicker).
;;
;; The colours are changed each interrupt so we have more than 4 colours on the screen
;; at the same time. 4 colours per region between interrupts.
;;
;; The sprite on this screen changes colours as it goes between the areas. 
;; This is expected. The sprite shares the palette with the screne.
;; 
;; 2 Ways to avoid this: 
;; 1 use the common set of 2 colours (2 colours for main sprite is not great), 
;; or make sure the sprite doesn't cross the regions
;;
;; In this example there are 2 colours that are common to each region: white and black
;; With 2 colours defined within each region. 

;; sprite width in bytes
sprite_width equ 64/4
sprite_height equ 64

;; firmware functions we use
scr_next_line equ &bc26
scr_set_mode equ &bc0e
scr_set_border equ &bc38
scr_set_ink equ &bc32
mc_wait_flyback equ &bd19
km_read_char equ &bb09
km_set_delay equ &bb3f
kl_new_fast_ticker equ &bce0
mc_set_mode        equ &bd1c
mc_set_inks        equ &bd25
kl_new_frame_fly   equ &bcd7
kl_init_event equ &bcef

;; table is aligned on a 256-byte boundary
mask_table equ &200

org &1000
nolist

start:
;; equivalent to SPEED KEY 1,1
;; tip from Markus to speed up key repeat 
;; when using km_read_char
ld h,1
ld l,1
call km_set_delay

;; set 4 colour mode 1
ld a,0
call scr_set_mode

;; set border to black
ld bc,0
call scr_set_border



;; wait for a screen refresh
;; we do this to synchronise our effect
call mc_wait_flyback

ld a,6
ld (ticker_counter),a
ld hl,colours
ld (current_colour_pointer),hl

;; install interrupt
ld hl,ticker_event_block
ld b,%10000010
ld c,&80
ld de,ticker_function
call kl_new_fast_ticker

ld hl,scr_pixels
ld de,&c000
ld bc,&4000
ldir

ret

call make_scr_table
call init_mask_table

;; define initial sprite coords
ld hl,50
ld (sprite_x),hl
ld a,100
ld (sprite_y),a

call calc_spr_scr_addr

call draw_sprite

;; the main loop which draws and updates sprite
main_loop:
call mc_wait_flyback


;; check keys and update sprite position
call check_keys

;; redraw sprite?
ld a,(spr_redraw)
or a
jr z,main_loop2

;; reset for next time
xor a
ld (spr_redraw),a

;; restore background (where sprite used to be)
ld hl,(spr_scr_addr)
ld de,sprite_background
ld b,sprite_height
ld c,sprite_width
;; low byte of x coordinate
ld a,(spr_pixel_offs)
call sprite_background_restore

;; recalc screen address for new position
call calc_spr_scr_addr

call draw_sprite

;; jumps to here if sprite has not moved and no update is required
main_loop2:

jp main_loop

draw_sprite:

;; store background pixels where sprite is located
ld hl,(spr_scr_addr)
ld de,sprite_background
ld b,sprite_height
ld c,sprite_width
;; low byte of x coordinate
ld a,(spr_pixel_offs)

call sprite_background_store


;; draw sprite (writes over background pixels)
ld de,sprite_pixel_ptrs
ld hl,(spr_scr_addr)
ld b,sprite_height
ld c,sprite_width

ld a,(spr_pixel_offs)


call sprite_draw
ret

;; This function will generate a lookup table of pixel masks.
;; This table can be used to plot masked sprites onto a mode 1 display.
;; 
;;
;; call init_mask_table to initialise.
;; 
;; 
;; NOTES:
;; - This code is for mode 1.
;; - The pen which will be plotted transparent is 0.
;; - Relocate mask_table to an address which is on a 256 byte boundary. (i.e. 
;; it's low byte is 0). This will allow you to access it much quicker.

init_mask_table:
ld hl,mask_table						;; base address of the mask table
ld b,0							;; counter
ld c,0							;; initial pixel value

mmt1:
ld d,0							;; initialise initial mask

ld a,c
and &88
jr z,mmt2
ld a,d
or &88
ld d,a

mmt2:
ld a,c
and &44
jr z,mmt3
ld a,d
or &44
ld d,a

mmt3:
ld a,c
and &22
jr z,mmt4
ld a,d
or &22
ld d,a

mmt4:
ld a,c
and &11
jr z,mmt5

ld a,d
or &11
ld d,a

mmt5:
ld a,d
cpl 
ld (hl),a						;; store final mask in table
inc hl							;; increment position in table
inc c							;; increment pixel value
djnz mmt1						;; loop
ret


;; check if a key has been pressed and perform action if it has
check_keys:
call km_read_char
ret nc
;; A = ascii char of key pressed
;; we check for both upper and lower case chars
cp 'q'
jp z,sprite_up
cp 'Q'
jp z,sprite_up
cp 'a'
jp z,sprite_down
cp 'A'
jp z,sprite_down
cp 'o'
jp z,sprite_left
cp 'O'
jp z,sprite_left
cp 'p'
jp z,sprite_right
cp 'P'
jp z,sprite_right
ret

;; move sprite one byte to left
;; 2 pixels in mode 0
sprite_left:
ld hl,(sprite_x)
ld a,h
or l
ret z
dec hl
ld (sprite_x),hl
ld a,1
ld (spr_redraw),a
ret

;; mode sprite one byte to right
;; 2 pixels in mode 0
sprite_width_pixels equ sprite_width*4
right_side equ 320-sprite_width_pixels

sprite_right:
ld hl,(sprite_x)
ld bc,right_side
or a
sbc hl,bc
ret p
add hl,bc
inc hl
ld (sprite_x),hl
ld a,1
ld (spr_redraw),a

ret

;; move sprite up one scan line
sprite_up:
ld a,(sprite_y)
or a
ret z
dec a
ld (sprite_y),a
ld a,1
ld (spr_redraw),a

ret

bottom_side equ 199-sprite_height

;; move sprite down one scan line
sprite_down:
ld a,(sprite_y)
cp bottom_side
ret z
inc a
ld (sprite_y),a
ld a,1
ld (spr_redraw),a

ret

;; Y line coordinate of sprite
sprite_y:
defb 0

;; x pixel coordinate of sprite
sprite_x:			
defw 0

;; screen address of sprite
spr_scr_addr:
defw 0

spr_pixel_offs:
defb 0

;; this will be 0 for no redraw
;; 1 otherwise
;; 
;; this is one way to reduce flicker by only
;; drawing sprite if it has moved
spr_redraw:
defb 0

;; From our x,y coordinates generate
;; a screen address for top-left of sprite to be drawn at.

;; IN:
;; H = x byte coord
;; L = y line
;; OUT:
;; HL = screen address
get_scr_addr:
push bc
push de
ld c,h
ld h,0
add hl,hl
ld de,scr_addr_table
add hl,de
ld a,(hl)
inc hl
ld h,(hl)
ld l,a
ld b,0
add hl,bc
pop de
pop bc
ret

;; generate table of screen addresses
;; one address per scanline. The address
;; is for the first byte of each line.
make_scr_table:
ld ix,scr_addr_table		;; address to store table
ld hl,&c000					;; start address of first scanline
ld b,200					;; number of scanlines on screen
mst1:
ld (ix+0),l
ld (ix+1),h
inc ix
inc ix
push bc
call scr_next_line
pop bc
djnz mst1
ret


calc_spr_scr_addr:
;; X screen coord in range 0..320
;; convert to byte coordinates by dividing X by 4
;; (4 = number of pixels per byte in mode 1)
ld hl,(sprite_x)
srl h				;; HL = HL / 2
rr l
srl h				;; HL = HL / 2
rr l
					;; result = HL = HL /4
ld h,l
;; now fetch y line coordinate
ld a,(sprite_y)
ld l,a
;; calculate screen address
call get_scr_addr
;; store it
ld (spr_scr_addr),hl

ld a,(sprite_x+0)
;; find pixel offset within byte (0..3)
and &3
ld (spr_pixel_offs),a
ret

;; HL = screen address
;; DE = pointer to array of sprite data
;; A = pixel offset in sprite
;; B = height
;; C = width

sprite_draw:
or a
jr z,sprite_draw2
inc c					;; when sprite is shifted, the width is one byte larger

sprite_draw2:
ex de,hl
add a,a				;; A = A * 2
add a,l				;; add low byte of array 
ld l,a
ld a,h	
adc a,0
ld h,a
;; HL = address in array
ld a,(hl)
inc hl
ld h,(hl)
ld l,a				;; read address from array = address of sprite pixel data
ex de,hl



push bc
pop ix							;; low byte of IX = C
								;; high byte of IX = B
								
;; B = upper 8-bits (bits 15..8) of mask table memory address
ld b,mask_table/256

sprite_draw_height:
push ix
push hl


sprite_draw_width:
ld a,(de)							;; get byte of sprite pixel data
ld c,a								;; C = byte of sprite pixel data/look-up table value
									;; BC = address (in look-up table) of mask corresponding to this sprite pixel data
ld a,(bc)							;; lookup mask from table
and (hl)							;; mask pixels on screen (remove pixels which will be replaced)
or c								;; combine with sprite pixel data
ld (hl),a							;; write result to screen 
inc de
inc hl
defb &dd							;; DEC LIX
dec l
jr nz,sprite_draw_width
pop hl
call scr_next_line
pop ix
defb &dd							;; DEC HIX
dec h
jr nz,sprite_draw_height
ret

;; HL = screen address of sprite
;; DE = address to store screen data
;; B = height
;; C = width
;; A = pixel offset into byte
;; store a rectangle from the screen into a buffer
sprite_background_store:
or a
jr z,sprite_back_height
inc c					;; when sprite is shifted, the width is one byte larger

sprite_back_height:
push bc
push hl

sprite_back_width:
ld a,(hl)				;; read from screen
ld (de),a				;; store to buffer
inc hl
inc de
dec c
jr nz,sprite_back_width

pop hl
call scr_next_line
pop bc
djnz sprite_back_height
ret

;; HL = screen address of sprite
;; H = x byte coord
;; L = y line
;; DE = address to store screen data
;; B = height
;; C = width
;; A = pixel offset into byte
;; restore a rectangle to the screen
sprite_background_restore:
or a
jr z,sprite_reback_height
inc c					;; when sprite is shifted, the width is one byte larger

sprite_reback_height:
push bc
push hl

sprite_reback_width:
ld a,(de)					;; read from buffer
ld (hl),a					;; write to screen
inc hl
inc de
dec c
jr nz,sprite_reback_width

pop hl
call scr_next_line
pop bc
djnz sprite_reback_height
ret

sprite_preshifted_width equ sprite_width+1
buffer_size equ sprite_height*sprite_preshifted_width

;; a buffer to store screen behind sprite
sprite_background:
defs buffer_size

;; table/array for screen addresses for each scan line
scr_addr_table:
defs 200*2

;; the inks for the pens
pens:
defb 1
defb 20
defb 2
defb 14


sprite_pixel_ptrs:
defw smiley1			;; sprite at pixel pos 0 in byte
defw smiley2			;; sprite at pixel pos 1 in byte
defw smiley3			;; sprite at pixel pos 2 in byte
defw smiley4			;; sprite at pixle pos 3 in byte


scr_pixels:
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055,&055
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff,&0ff
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0,&0f0
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5,&0a5
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f,&00f
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
defb &000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000,&000
;; scr

smiley1:
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&10,&fe,&80,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&31,&f9,&c0,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&d0,&00,&30,&80,&60,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&70,&fe,&80,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&f3,&f1,&40,&0f,&0e,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&10,&e5,&0f,&87,&0f,&e1,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&16,&87,&0f,&0f,&1e,&c3,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&01,&1e,&0f,&0f,&0f,&1e,&87,&0f,&00,&00,&00,&00,&00 
defb &00,&00,&00,&03,&0f,&0f,&0f,&0f,&1e,&87,&0f,&0c,&00,&00,&00,&00 
defb &00,&00,&01,&0f,&0f,&0f,&0f,&0f,&1f,&87,&0f,&0e,&00,&00,&00,&00 
defb &00,&00,&03,&0f,&0f,&0f,&3c,&f0,&f0,&87,&0f,&0f,&00,&00,&00,&00 
defb &00,&10,&03,&0f,&0f,&0f,&7a,&96,&f3,&87,&0f,&0f,&08,&00,&00,&00 
defb &00,&03,&07,&0f,&0f,&0f,&e5,&0f,&1e,&8f,&0f,&0f,&08,&00,&00,&00 
defb &00,&25,&0f,&0f,&0f,&1e,&e9,&0f,&1e,&87,&0f,&0f,&0c,&00,&00,&00 
defb &00,&4b,&0f,&0f,&0f,&3c,&cb,&0f,&0f,&87,&0f,&0f,&0e,&00,&00,&00 
defb &00,&c3,&0f,&0f,&0f,&3d,&87,&0f,&0f,&0f,&0f,&0f,&0f,&00,&00,&00 
defb &00,&87,&0f,&0f,&0f,&3c,&87,&0f,&78,&0f,&e5,&0f,&0f,&08,&00,&00 
defb &10,&87,&0f,&0f,&0f,&3e,&0f,&0f,&f7,&cb,&f6,&0f,&0f,&08,&00,&00 
defb &10,&0f,&0f,&0f,&0f,&7a,&0f,&1e,&ff,&cf,&f6,&0f,&0f,&0c,&00,&00 
defb &30,&0f,&0f,&0f,&0f,&78,&0f,&1e,&ff,&ed,&69,&0f,&0f,&0e,&00,&00 
defb &30,&0f,&78,&87,&0f,&69,&0f,&1e,&f7,&ed,&0f,&0f,&0f,&0e,&00,&00 
defb &61,&0f,&f0,&c3,&0f,&69,&0f,&0f,&f7,&ed,&0f,&0f,&0f,&0e,&00,&00 
defb &61,&0f,&f0,&c3,&0f,&78,&0f,&0f,&f5,&ed,&0f,&0f,&0f,&0e,&00,&00 
defb &61,&0f,&f0,&c3,&0f,&6b,&0f,&0f,&78,&c3,&4b,&0f,&0f,&0e,&00,&00 
defb &e1,&0f,&f0,&c3,&0f,&7a,&0f,&0f,&0f,&0f,&4f,&0f,&0f,&0e,&00,&00 
defb &e1,&87,&78,&c3,&0f,&7a,&0f,&0f,&0f,&0f,&e7,&0f,&0f,&0e,&00,&00 
defb &e1,&0f,&78,&87,&0f,&3d,&0f,&0f,&0f,&0f,&fd,&87,&0f,&2c,&00,&00 
defb &f0,&4b,&0f,&0f,&0f,&3d,&87,&0f,&0f,&1e,&f8,&cb,&0f,&2c,&00,&00 
defb &e1,&87,&3c,&0f,&0f,&1e,&cb,&0f,&0f,&1f,&da,&f4,&87,&68,&00,&00 
defb &f8,&4b,&78,&87,&0f,&1e,&e5,&0f,&0f,&3f,&87,&7b,&fc,&e8,&00,&00 
defb &f0,&4b,&3c,&87,&0f,&0f,&7a,&87,&0f,&f2,&0f,&3c,&f2,&e0,&00,&00 
defb &f8,&a5,&3c,&0f,&0f,&69,&3d,&fc,&f1,&e9,&1e,&4b,&d2,&e0,&00,&00 
defb &f0,&a5,&87,&0f,&0f,&e1,&0f,&f3,&f6,&87,&0f,&3c,&d2,&68,&00,&00 
defb &f8,&d2,&0f,&0f,&0f,&f0,&0f,&1e,&4b,&0f,&5a,&5a,&b4,&e0,&00,&00 
defb &74,&e1,&a5,&0f,&0f,&e1,&0f,&0f,&0f,&0f,&0f,&b4,&d2,&e0,&00,&00 
defb &74,&e1,&87,&0f,&0f,&4b,&0f,&0f,&0f,&1e,&5a,&5a,&f0,&e0,&00,&00 
defb &74,&f0,&5a,&0f,&1e,&87,&0f,&0f,&0f,&0f,&2d,&b4,&d2,&e0,&00,&00 
defb &74,&f0,&a5,&87,&3c,&e1,&0f,&0f,&0f,&5a,&5a,&5a,&f0,&c0,&00,&00 
defb &72,&f0,&d2,&5a,&b4,&e1,&0f,&0f,&0f,&0f,&a5,&b4,&f0,&c0,&00,&00 
defb &72,&f0,&e1,&87,&3c,&f0,&0f,&0f,&a5,&a5,&a5,&d2,&f0,&c0,&00,&00 
defb &32,&f0,&f0,&69,&78,&e1,&0f,&2d,&0f,&1e,&5a,&78,&f0,&f0,&e0,&00 
defb &31,&f0,&f0,&f0,&78,&e1,&a5,&1e,&5a,&5a,&78,&f0,&f1,&ff,&ff,&c0 
defb &31,&f0,&f0,&f0,&3c,&e1,&5a,&4b,&a5,&a5,&b4,&f0,&f3,&ff,&ff,&ce 
defb &12,&f8,&f0,&d2,&f0,&f0,&5a,&5a,&f0,&f0,&f0,&f0,&ff,&ff,&ff,&ef 
defb &10,&f8,&f0,&f0,&f0,&f0,&f0,&f0,&f0,&78,&f0,&f1,&ff,&ff,&fe,&fa 
defb &01,&f8,&f0,&f0,&f0,&e1,&a5,&a5,&b4,&f0,&b4,&f3,&ff,&ff,&fa,&fb 
defb &00,&f4,&f0,&f0,&f0,&f0,&5a,&5a,&f1,&ff,&f8,&ff,&ff,&ff,&fd,&f1 
defb &00,&f6,&f0,&f0,&f0,&d2,&f0,&e1,&f3,&f1,&ff,&ff,&ff,&f4,&f4,&f1 
defb &00,&73,&f0,&f0,&f0,&f0,&f0,&f0,&f6,&f0,&f1,&f7,&fe,&fa,&f0,&f1 
defb &00,&71,&f0,&f0,&f0,&f0,&f0,&f8,&fc,&f0,&f0,&f3,&ff,&f0,&f0,&f1 
defb &00,&30,&f8,&f0,&f0,&f0,&f0,&f7,&f0,&f0,&f0,&f0,&f1,&f8,&f0,&f1 
defb &00,&10,&f6,&f0,&f0,&f0,&f0,&f0,&f0,&f0,&f0,&e1,&10,&f6,&f0,&f2 
defb &00,&00,&73,&f0,&f0,&f0,&f0,&f0,&f0,&f0,&f0,&c0,&00,&71,&f0,&e0 
defb &00,&00,&31,&fc,&f0,&f0,&f0,&f0,&f0,&f0,&f0,&80,&00,&10,&f5,&c0 
defb &00,&00,&10,&fe,&f0,&f0,&f0,&f0,&f0,&f0,&f4,&00,&00,&00,&70,&80 
defb &00,&00,&00,&f7,&fe,&f0,&f0,&f0,&f0,&f1,&c0,&00,&00,&00,&00,&00 
defb &00,&00,&00,&11,&ff,&f8,&f0,&f0,&f1,&cc,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&33,&ff,&fe,&fb,&ff,&c8,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&77,&ff,&ff,&fe,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00
;; smiley1

smiley2:
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&f7,&c0,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&10,&fc,&e8,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&60,&80,&10,&c0,&30,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&30,&f7,&c0,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&71,&f8,&a8,&07,&0f,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&f2,&0f,&4b,&0f,&78 
defb &08,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&03,&c3,&0f,&0f,&0f 
defb &e1,&08,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&0f,&87,&0f,&0f 
defb &0f,&c3,&0f,&08,&00,&00,&00,&00,&00,&00,&00,&00,&01,&0f,&0f,&0f 
defb &0f,&0f,&c3,&0f,&0e,&00,&00,&00,&00,&00,&00,&00,&00,&0f,&0f,&0f 
defb &0f,&0f,&0f,&cb,&0f,&0f,&00,&00,&00,&00,&00,&00,&00,&01,&0f,&0f 
defb &0f,&1e,&f0,&f0,&c3,&0f,&0f,&08,&00,&00,&00,&00,&00,&00,&81,&0f 
defb &0f,&0f,&3d,&c3,&f1,&cb,&0f,&0f,&0c,&00,&00,&00,&00,&00,&01,&0b 
defb &0f,&0f,&0f,&7a,&0f,&0f,&c7,&0f,&0f,&0c,&00,&00,&00,&00,&00,&12 
defb &0f,&0f,&0f,&0f,&f4,&0f,&0f,&c3,&0f,&0f,&0e,&00,&00,&00,&00,&00 
defb &25,&0f,&0f,&0f,&1e,&e5,&0f,&0f,&4b,&0f,&0f,&0f,&00,&00,&00,&00 
defb &00,&61,&0f,&0f,&0f,&1e,&cb,&0f,&0f,&0f,&0f,&0f,&0f,&08,&00,&00 
defb &00,&00,&43,&0f,&0f,&0f,&1e,&c3,&0f,&3c,&87,&7a,&0f,&0f,&0c,&00 
defb &00,&00,&00,&c3,&0f,&0f,&0f,&1f,&87,&0f,&7b,&ed,&7b,&87,&0f,&0c 
defb &00,&00,&00,&00,&87,&0f,&0f,&0f,&3d,&87,&0f,&f7,&ef,&7b,&87,&0f 
defb &0e,&00,&00,&00,&10,&87,&0f,&0f,&0f,&3c,&87,&0f,&f7,&fe,&3c,&0f 
defb &0f,&0f,&00,&00,&00,&10,&87,&3c,&c3,&0f,&3c,&0f,&0f,&f3,&fe,&0f 
defb &0f,&0f,&0f,&00,&00,&00,&30,&0f,&78,&e1,&0f,&3c,&0f,&0f,&7b,&fe 
defb &0f,&0f,&0f,&0f,&00,&00,&00,&30,&0f,&78,&e1,&0f,&3c,&87,&0f,&7a 
defb &fe,&0f,&0f,&0f,&0f,&00,&00,&00,&30,&0f,&78,&e1,&0f,&3d,&0f,&0f 
defb &3c,&e1,&2d,&0f,&0f,&0f,&00,&00,&00,&70,&0f,&78,&e1,&0f,&3d,&87 
defb &0f,&0f,&0f,&2f,&0f,&0f,&0f,&00,&00,&00,&70,&4b,&3c,&e1,&0f,&3d 
defb &87,&0f,&0f,&0f,&7b,&0f,&0f,&0f,&00,&00,&00,&70,&0f,&3c,&c3,&0f 
defb &1e,&8f,&0f,&0f,&0f,&7e,&cb,&0f,&1e,&00,&00,&00,&70,&a5,&0f,&0f 
defb &0f,&1e,&cb,&0f,&0f,&0f,&f4,&e5,&0f,&1e,&00,&00,&00,&70,&4b,&1e 
defb &87,&0f,&0f,&e5,&0f,&0f,&0f,&ed,&f2,&c3,&3c,&00,&00,&00,&74,&a5 
defb &3c,&c3,&0f,&0f,&f2,&0f,&0f,&1f,&cb,&3d,&fe,&f4,&00,&00,&00,&70 
defb &a5,&1e,&c3,&0f,&0f,&3d,&c3,&0f,&79,&87,&1e,&f1,&f0,&00,&00,&00 
defb &74,&d2,&1e,&87,&0f,&3c,&1e,&fe,&f0,&fc,&0f,&a5,&69,&f0,&00,&00 
defb &00,&70,&d2,&4b,&0f,&0f,&78,&0f,&79,&fb,&c3,&0f,&1e,&e1,&b4,&00 
defb &00,&00,&74,&e1,&87,&0f,&0f,&78,&87,&0f,&a5,&0f,&2d,&a5,&d2,&f0 
defb &00,&00,&00,&32,&f0,&5a,&0f,&0f,&78,&0f,&0f,&0f,&0f,&0f,&5a,&e1 
defb &f0,&00,&00,&00,&32,&f0,&4b,&0f,&0f,&2d,&0f,&0f,&0f,&0f,&a5,&a5 
defb &f0,&f0,&00,&00,&00,&32,&f0,&a5,&87,&0f,&c3,&0f,&0f,&0f,&0f,&1e 
defb &5a,&e1,&f0,&00,&00,&00,&32,&f0,&d2,&4b,&1e,&f0,&0f,&0f,&0f,&2d 
defb &a5,&a5,&f0,&e0,&00,&00,&00,&31,&f0,&e1,&a5,&d2,&f0,&0f,&0f,&0f 
defb &0f,&5a,&5a,&f0,&e0,&00,&00,&00,&31,&f0,&f0,&4b,&1e,&f0,&87,&0f 
defb &5a,&5a,&5a,&69,&f0,&e0,&00,&00,&00,&11,&f0,&f0,&b4,&3c,&f0,&0f 
defb &1e,&0f,&0f,&a5,&b4,&f0,&f0,&f0,&00,&00,&10,&f8,&f0,&f0,&b4,&f0 
defb &5a,&0f,&a5,&a5,&b4,&f0,&f0,&ff,&ff,&e8,&00,&10,&f8,&f0,&f0,&96 
defb &f0,&2d,&a5,&5a,&5a,&5a,&f0,&f1,&ff,&ff,&ef,&00,&01,&f4,&f0,&e1 
defb &f0,&f0,&a5,&a5,&f0,&f0,&f0,&f0,&f7,&ff,&ff,&ff,&08,&00,&f4,&f0 
defb &f0,&f0,&f0,&f0,&f0,&f0,&b4,&f0,&f0,&ff,&ff,&ff,&f5,&80,&00,&7c 
defb &f0,&f0,&f0,&f0,&5a,&5a,&5a,&f0,&d2,&f1,&ff,&ff,&fd,&f5,&88,&00 
defb &72,&f0,&f0,&f0,&f0,&a5,&a5,&f0,&ff,&fc,&f7,&ff,&ff,&fe,&f8,&88 
defb &00,&73,&f0,&f0,&f0,&e1,&f0,&f0,&79,&f8,&ff,&ff,&ff,&fa,&f2,&f0 
defb &88,&00,&31,&f8,&f0,&f0,&f0,&f0,&f0,&f3,&f0,&f0,&fb,&ff,&f5,&f0 
defb &f0,&88,&00,&30,&f8,&f0,&f0,&f0,&f0,&f4,&f6,&f0,&f0,&f1,&ff,&f8 
defb &f0,&f0,&88,&00,&10,&f4,&f0,&f0,&f0,&f0,&f3,&f8,&f0,&f0,&f0,&f0 
defb &fc,&f0,&f0,&88,&00,&00,&f3,&f0,&f0,&f0,&f0,&f0,&f0,&f0,&f0,&f0 
defb &08,&f3,&f0,&f1,&80,&00,&00,&31,&f8,&f0,&f0,&f0,&f0,&f0,&f0,&f0 
defb &e0,&00,&30,&f8,&f0,&00,&00,&00,&10,&fe,&f0,&f0,&f0,&f0,&f0,&f0 
defb &f0,&c0,&00,&00,&f2,&e8,&00,&00,&00,&00,&f7,&f0,&f0,&f0,&f0,&f0 
defb &f0,&f2,&80,&00,&00,&30,&c0,&00,&00,&00,&00,&73,&ff,&f0,&f0,&f0 
defb &f0,&f0,&e8,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&ff,&fc,&f0 
defb &f0,&f0,&ee,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&11,&ff 
defb &ff,&f5,&ff,&ec,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &33,&ff,&ff,&ff,&80,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00
;; smiley2

smiley3:
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&73,&e8,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&f6,&f4,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&30,&40,&00,&e0,&10,&80,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&10,&f3,&e8,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&30,&fc,&d4,&03,&0f,&08 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&71,&87,&2d,&0f,&3c 
defb &84,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&01,&69,&0f,&0f,&0f 
defb &78,&0c,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&07,&4b,&0f,&0f 
defb &0f,&69,&0f,&0c,&00,&00,&00,&00,&00,&00,&00,&00,&00,&0f,&0f,&0f 
defb &0f,&0f,&69,&0f,&0f,&00,&00,&00,&00,&00,&00,&00,&00,&07,&0f,&0f 
defb &0f,&0f,&0f,&6d,&0f,&0f,&08,&00,&00,&00,&00,&00,&00,&00,&0f,&0f 
defb &0f,&0f,&f0,&f0,&e1,&0f,&0f,&0c,&00,&00,&00,&00,&00,&00,&40,&0f 
defb &0f,&0f,&1e,&e9,&78,&ed,&0f,&0f,&0e,&00,&00,&00,&00,&00,&00,&0d 
defb &0f,&0f,&0f,&3d,&87,&0f,&6b,&0f,&0f,&0e,&00,&00,&00,&00,&00,&01 
defb &87,&0f,&0f,&0f,&7a,&87,&0f,&69,&0f,&0f,&0f,&00,&00,&00,&00,&00 
defb &12,&0f,&0f,&0f,&0f,&f2,&0f,&0f,&2d,&0f,&0f,&0f,&08,&00,&00,&00 
defb &00,&30,&0f,&0f,&0f,&0f,&e5,&0f,&0f,&0f,&0f,&0f,&0f,&0c,&00,&00 
defb &00,&00,&21,&0f,&0f,&0f,&0f,&e1,&0f,&1e,&c3,&3d,&87,&0f,&0e,&00 
defb &00,&00,&00,&61,&0f,&0f,&0f,&0f,&cb,&0f,&3d,&fe,&3d,&cb,&0f,&0e 
defb &00,&00,&00,&00,&43,&0f,&0f,&0f,&1e,&cb,&0f,&7b,&ff,&3d,&cb,&0f 
defb &0f,&00,&00,&00,&00,&c3,&0f,&0f,&0f,&1e,&c3,&0f,&7b,&ff,&96,&87 
defb &0f,&0f,&08,&00,&00,&00,&c3,&1e,&e1,&0f,&1e,&87,&0f,&79,&ff,&87 
defb &0f,&0f,&0f,&08,&00,&00,&10,&87,&3c,&f0,&0f,&1e,&87,&0f,&3d,&ff 
defb &87,&0f,&0f,&0f,&08,&00,&00,&10,&87,&3c,&f0,&0f,&1e,&c3,&0f,&3d 
defb &f7,&87,&0f,&0f,&0f,&08,&00,&00,&10,&87,&3c,&f0,&0f,&1e,&8f,&0f 
defb &1e,&f0,&1e,&0f,&0f,&0f,&08,&00,&00,&30,&87,&3c,&f0,&0f,&1e,&cb 
defb &0f,&0f,&0f,&1f,&0f,&0f,&0f,&08,&00,&00,&30,&a5,&1e,&f0,&0f,&1e 
defb &cb,&0f,&0f,&0f,&3d,&8f,&0f,&0f,&08,&00,&00,&30,&87,&1e,&e1,&0f 
defb &0f,&c7,&0f,&0f,&0f,&3f,&e5,&0f,&0f,&80,&00,&00,&30,&d2,&0f,&0f 
defb &0f,&0f,&e5,&0f,&0f,&0f,&7a,&f2,&0f,&0f,&80,&00,&00,&30,&a5,&0f 
defb &c3,&0f,&0f,&7a,&0f,&0f,&0f,&7e,&79,&e1,&1e,&80,&00,&00,&32,&d2 
defb &1e,&e1,&0f,&0f,&79,&87,&0f,&0f,&ed,&1e,&ff,&f2,&80,&00,&00,&30 
defb &d2,&0f,&e1,&0f,&0f,&1e,&e9,&0f,&3c,&cb,&0f,&f0,&f8,&80,&00,&00 
defb &32,&e1,&87,&c3,&0f,&1e,&87,&f7,&f0,&f6,&87,&5a,&3c,&78,&80,&00 
defb &00,&30,&e1,&a5,&0f,&0f,&3c,&87,&3c,&fd,&e9,&0f,&0f,&f0,&5a,&80 
defb &00,&00,&32,&f0,&4b,&0f,&0f,&3c,&c3,&0f,&5a,&0f,&1e,&5a,&69,&f0 
defb &80,&00,&00,&11,&f0,&a5,&87,&0f,&3c,&87,&0f,&0f,&0f,&0f,&2d,&f0 
defb &78,&80,&00,&00,&11,&f0,&a5,&0f,&0f,&1e,&0f,&0f,&0f,&0f,&5a,&5a 
defb &78,&f0,&80,&00,&00,&11,&f0,&d2,&4b,&0f,&69,&0f,&0f,&0f,&0f,&0f 
defb &a5,&f0,&78,&80,&00,&00,&11,&f0,&e1,&a5,&0f,&f0,&87,&0f,&0f,&1e 
defb &5a,&5a,&78,&f0,&00,&00,&00,&10,&f8,&f0,&5a,&69,&f0,&87,&0f,&0f 
defb &0f,&2d,&a5,&f0,&f0,&00,&00,&00,&10,&f8,&f0,&a5,&0f,&f0,&c3,&0f 
defb &2d,&a5,&a5,&b4,&78,&f0,&00,&00,&00,&00,&f8,&f0,&d2,&96,&f0,&87 
defb &0f,&87,&0f,&5a,&5a,&f0,&f0,&f0,&80,&00,&00,&f4,&f0,&f0,&d2,&f0 
defb &a5,&87,&5a,&5a,&5a,&f0,&f0,&f7,&ff,&fc,&00,&00,&f4,&f0,&f0,&c3 
defb &f0,&96,&5a,&2d,&a5,&a5,&f0,&f0,&ff,&ff,&ff,&08,&00,&7a,&f0,&f0 
defb &78,&f0,&d2,&5a,&78,&f0,&f0,&f0,&f3,&ff,&ff,&ff,&8c,&00,&72,&f0 
defb &f0,&f0,&f0,&f0,&f0,&f0,&d2,&f0,&f0,&f7,&ff,&ff,&fa,&c8,&00,&36 
defb &f0,&f0,&f0,&f0,&a5,&a5,&a5,&f0,&e1,&f0,&ff,&ff,&fe,&fa,&cc,&00 
defb &31,&f0,&f0,&f0,&f0,&d2,&5a,&78,&f7,&fe,&f3,&ff,&ff,&ff,&f4,&c4 
defb &00,&31,&f8,&f0,&f0,&f0,&78,&f0,&b4,&fc,&f7,&ff,&ff,&fd,&f1,&f0 
defb &c4,&00,&10,&fc,&f0,&f0,&f0,&f0,&f0,&f1,&f8,&f0,&f5,&ff,&fa,&f8 
defb &f0,&c4,&00,&10,&f4,&f0,&f0,&f0,&f0,&f2,&f3,&f0,&f0,&f0,&ff,&fc 
defb &f0,&f0,&c4,&00,&00,&f2,&f0,&f0,&f0,&f0,&f1,&fc,&f0,&f0,&f0,&f0 
defb &f6,&f0,&f0,&c4,&00,&00,&71,&f8,&f0,&f0,&f0,&f0,&f0,&f0,&f0,&f0 
defb &84,&71,&f8,&f0,&c8,&00,&00,&10,&fc,&f0,&f0,&f0,&f0,&f0,&f0,&f0 
defb &f0,&00,&10,&f4,&f0,&80,&00,&00,&00,&f7,&f0,&f0,&f0,&f0,&f0,&f0 
defb &f0,&e0,&00,&00,&71,&f4,&00,&00,&00,&00,&73,&f8,&f0,&f0,&f0,&f0 
defb &f0,&f1,&c0,&00,&00,&10,&e0,&00,&00,&00,&00,&31,&ff,&f8,&f0,&f0 
defb &f0,&f0,&f4,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&77,&fe,&f0 
defb &f0,&f0,&f7,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&ff 
defb &ff,&fa,&ff,&fe,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &11,&ff,&ff,&ff,&c8,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00
;; smiley3

smiley4:
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&31,&fc,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&73,&f2,&80,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&10,&a0,&00,&70,&00,&c0,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&f1,&fc,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&10,&f6,&e2,&81,&0f,&0c 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&30,&cb,&1e,&0f,&1e 
defb &c2,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&3c,&0f,&0f,&0f 
defb &3c,&86,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&03,&2d,&0f,&0f 
defb &0f,&3c,&0f,&0e,&00,&00,&00,&00,&00,&00,&00,&00,&00,&07,&0f,&0f 
defb &0f,&0f,&3c,&0f,&0f,&08,&00,&00,&00,&00,&00,&00,&00,&03,&0f,&0f 
defb &0f,&0f,&0f,&3e,&0f,&0f,&0c,&00,&00,&00,&00,&00,&00,&00,&07,&0f 
defb &0f,&0f,&78,&f0,&f0,&0f,&0f,&0e,&00,&00,&00,&00,&00,&00,&20,&07 
defb &0f,&0f,&0f,&f4,&3c,&f6,&0f,&0f,&0f,&00,&00,&00,&00,&00,&00,&06 
defb &0f,&0f,&0f,&1e,&cb,&0f,&3d,&0f,&0f,&0f,&00,&00,&00,&00,&00,&00 
defb &4b,&0f,&0f,&0f,&3d,&c3,&0f,&3c,&0f,&0f,&0f,&08,&00,&00,&00,&00 
defb &01,&87,&0f,&0f,&0f,&79,&87,&0f,&1e,&0f,&0f,&0f,&0c,&00,&00,&00 
defb &00,&10,&87,&0f,&0f,&0f,&7a,&0f,&0f,&0f,&0f,&0f,&0f,&0e,&00,&00 
defb &00,&00,&10,&0f,&0f,&0f,&0f,&78,&0f,&0f,&e1,&1e,&cb,&0f,&0f,&00 
defb &00,&00,&00,&30,&0f,&0f,&0f,&0f,&6d,&0f,&1e,&ff,&96,&ed,&0f,&0f 
defb &00,&00,&00,&00,&21,&0f,&0f,&0f,&0f,&e5,&0f,&3d,&ff,&9e,&ed,&0f 
defb &0f,&08,&00,&00,&00,&61,&0f,&0f,&0f,&0f,&e1,&0f,&3d,&ff,&cb,&c3 
defb &0f,&0f,&0c,&00,&00,&00,&61,&0f,&f0,&0f,&0f,&c3,&0f,&3c,&ff,&cb 
defb &0f,&0f,&0f,&0c,&00,&00,&00,&c3,&1e,&f0,&87,&0f,&c3,&0f,&1e,&ff 
defb &cb,&0f,&0f,&0f,&0c,&00,&00,&00,&c3,&1e,&f0,&87,&0f,&e1,&0f,&1e 
defb &fb,&cb,&0f,&0f,&0f,&0c,&00,&00,&00,&c3,&1e,&f0,&87,&0f,&c7,&0f 
defb &0f,&f0,&87,&87,&0f,&0f,&0c,&00,&00,&10,&c3,&1e,&f0,&87,&0f,&e5 
defb &0f,&0f,&0f,&0f,&8f,&0f,&0f,&0c,&00,&00,&10,&d2,&0f,&f0,&87,&0f 
defb &e5,&0f,&0f,&0f,&1e,&cf,&0f,&0f,&0c,&00,&00,&10,&c3,&0f,&f0,&0f 
defb &0f,&6b,&0f,&0f,&0f,&1f,&fa,&0f,&0f,&48,&00,&00,&10,&e1,&87,&0f 
defb &0f,&0f,&7a,&0f,&0f,&0f,&3d,&f1,&87,&0f,&48,&00,&00,&10,&d2,&0f 
defb &69,&0f,&0f,&3d,&87,&0f,&0f,&3f,&b4,&f8,&0f,&c0,&00,&00,&11,&e1 
defb &87,&f0,&0f,&0f,&3c,&cb,&0f,&0f,&7e,&0f,&f7,&f9,&c0,&00,&00,&10 
defb &e1,&87,&78,&0f,&0f,&0f,&f4,&0f,&1e,&e5,&0f,&78,&f4,&c0,&00,&00 
defb &11,&f0,&4b,&69,&0f,&0f,&c3,&7b,&f8,&f3,&c3,&2d,&96,&b4,&c0,&00 
defb &00,&10,&f0,&5a,&0f,&0f,&1e,&c3,&1e,&f6,&fc,&0f,&0f,&78,&a5,&c0 
defb &00,&00,&11,&f0,&a5,&0f,&0f,&1e,&e1,&0f,&2d,&87,&0f,&a5,&b4,&78 
defb &c0,&00,&00,&00,&f8,&d2,&4b,&0f,&1e,&c3,&0f,&0f,&0f,&0f,&1e,&78 
defb &b4,&c0,&00,&00,&00,&f8,&d2,&0f,&0f,&0f,&87,&0f,&0f,&0f,&2d,&a5 
defb &b4,&f0,&c0,&00,&00,&00,&f8,&e1,&a5,&0f,&3c,&0f,&0f,&0f,&0f,&0f 
defb &5a,&78,&b4,&c0,&00,&00,&00,&f8,&f0,&5a,&0f,&78,&c3,&0f,&0f,&0f 
defb &a5,&a5,&b4,&f0,&80,&00,&00,&00,&f4,&f0,&a5,&b4,&78,&c3,&0f,&0f 
defb &0f,&1e,&5a,&78,&f0,&80,&00,&00,&00,&f4,&f0,&d2,&0f,&78,&e1,&0f 
defb &1e,&5a,&5a,&5a,&b4,&f0,&80,&00,&00,&00,&74,&f0,&e1,&c3,&f0,&c3 
defb &0f,&4b,&0f,&2d,&a5,&f0,&f0,&f0,&c0,&00,&00,&72,&f0,&f0,&e1,&f0 
defb &d2,&4b,&2d,&a5,&a5,&f0,&f0,&f3,&ff,&fe,&80,&00,&72,&f0,&f0,&e1 
defb &78,&c3,&a5,&96,&5a,&5a,&78,&f0,&f7,&ff,&ff,&8c,&00,&35,&f0,&f0 
defb &b4,&f0,&e1,&a5,&b4,&f0,&f0,&f0,&f1,&ff,&ff,&ff,&ce,&00,&31,&f0 
defb &f0,&f0,&f0,&f0,&f0,&f0,&e1,&f0,&f0,&f3,&ff,&ff,&fd,&e4,&00,&13 
defb &f0,&f0,&f0,&f0,&d2,&5a,&5a,&78,&f0,&78,&f7,&ff,&ff,&f5,&e6,&00 
defb &10,&f8,&f0,&f0,&f0,&e1,&a5,&b4,&f3,&ff,&f1,&ff,&ff,&ff,&fa,&e2 
defb &00,&10,&fc,&f0,&f0,&f0,&b4,&f0,&d2,&f6,&f3,&ff,&ff,&fe,&f8,&f8 
defb &e2,&00,&00,&f6,&f0,&f0,&f0,&f0,&f0,&f0,&fc,&f0,&f2,&ff,&fd,&f4 
defb &f0,&e2,&00,&00,&f2,&f0,&f0,&f0,&f0,&f1,&f1,&f8,&f0,&f0,&f7,&fe 
defb &f0,&f0,&e2,&00,&00,&71,&f0,&f0,&f0,&f0,&f0,&fe,&f0,&f0,&f0,&f0 
defb &f3,&f0,&f0,&e2,&00,&00,&30,&fc,&f0,&f0,&f0,&f0,&f0,&f0,&f0,&f0 
defb &c2,&30,&fc,&f0,&e4,&00,&00,&00,&f6,&f0,&f0,&f0,&f0,&f0,&f0,&f0 
defb &f0,&80,&00,&f2,&f0,&c0,&00,&00,&00,&73,&f8,&f0,&f0,&f0,&f0,&f0 
defb &f0,&f0,&00,&00,&30,&fa,&80,&00,&00,&00,&31,&fc,&f0,&f0,&f0,&f0 
defb &f0,&f0,&e8,&00,&00,&00,&f0,&00,&00,&00,&00,&10,&ff,&fc,&f0,&f0 
defb &f0,&f0,&f2,&80,&00,&00,&00,&00,&00,&00,&00,&00,&00,&33,&ff,&f0 
defb &f0,&f0,&f3,&88,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&77 
defb &ff,&fd,&f7,&ff,&80,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&ff,&ff,&ff,&ec,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00 
defb &00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00,&00
;; smiley4


;; this is initialised by
;; the firmware; holds runtime state of ticker interrupt
ticker_event_block:
defs 10

;; this is the function called each 1/300th of a second
ticker_function:
push af
push hl

;; The 1/300th of a second interrupt effectively splits
;; the screen into 6 sections of equal height. Each section
;; spans the entire width of the screen.
;;
;; We want to ensure that the effect is stationary so we reset
;; every 6 calls of this function.
ld a,(ticker_counter)
dec a
ld (ticker_counter),a
or a
jr nz,ticker_function2
ld a,6
ld (ticker_counter),a
ld hl,colours
ld (current_colour_pointer),hl

ticker_function2:

;; setting the colours will occur immeditately.

;; get pointer to current colours
ld de,(current_colour_pointer)
call mc_set_inks

;; update colours pointer
ld hl,(current_colour_pointer)
ld bc,17
add hl,bc
ld (current_colour_pointer),hl


pop af
pop hl
ret

ticker_counter: defb 0

current_colour_pointer: defw colours


;; The 1/300th of a second interrupt effectively splits
;; the screen into 6 sections of equal height. Each section
;; spans the entire width of the screen.
;;
;; video mode for each of the 6 sections of the screen.

colours:

;; colours for 1st section
;; border colour, followed by pen 0, pen 1 up to pen 15
;; NOTE: These are hardware colour numbers
;; NOTE: We are using mode 1, so we only change the first 4 colours.
; BORDER, PAPER 1,

; HARDWARE COLOURS
;&54,&44,&55,&5C,&58,&5D,&4C,&45,&4D,&56
;&46,&57,&5E,&40,&F5,&4E,&47,&4F,&52,&42
;&53,&5A,&59,&5B,&4A,&43,&4B

defb &04,&04,&54,&44,&55,&5C,&58,&5D,&4C,&45,&4D,&56,&46,&57,&5E,&40,&F5
defb &04,&04,&54,&44,&55,&5C,&58,&5D,&4C,&45,&4D,&56,&46,&57,&5E,&40,&F5
defb &04,&04,&54,&44,&55,&5C,&58,&5D,&4C,&45,&4D,&56,&46,&57,&5E,&40,&F5
defb &04,&04,&5E,&40,&F5,&4E,&47,&4F,&52,&42,&53,&5A,&59,&5B,&4A,&43,&4B
defb &04,&04,&5E,&40,&F5,&4E,&47,&4F,&52,&42,&53,&5A,&59,&5B,&4A,&43,&4B
defb &04,&04,&5E,&40,&F5,&4E,&47,&4F,&52,&42,&53,&5A,&59,&5B,&4A,&43,&4B
end 